/******************************************************************************
 *
 * Copyright (c) 1995-2002 Palm, Inc. or its subsidiaries.
 * All rights reserved.
 *
 * File: ToDoDB.c
 *
 * Release: Palm OS SDK 5.0 (111823)
 *
 * Description:
 *		To Do Manager routines
 *
 * History:
 *   	1/20/95  rsf - Created
 *		10/4/99	jmp - Added ToDoGetDatabase() routine.
 *		10/4/99	rbb - Moved SetDBBackupBit from ToDo.c (used by
 *							ToDoGetDatabase) and renamed as ToDoSetDBBackupBit
 *
 *****************************************************************************/

#include <PalmOS.h>
#include "ToDoDB.h"


/************************************************************
 *
 *  FUNCTION: DateTypeCmp
 *
 *  DESCRIPTION: Compare two dates
 *
 *  PARAMETERS: 
 *
 *  RETURNS: 
 *
 *  CREATED: 1/20/95 
 *
 *  BY: Roger Flores
 *
 *************************************************************/
static Int16 DateTypeCmp(DateType d1, DateType d2)
{
	Int16 result;
	
	result = d1.year - d2.year;
	if (result != 0)
		{
		if (DateToInt(d1) == 0xffff)
			return 1;
		if (DateToInt(d2) == 0xffff)
			return -1;
		return result;
		}
	
	result = d1.month - d2.month;
	if (result != 0)
		return result;
	
	result = d1.day - d2.day;

	return result;

}


/************************************************************
 *
 *  FUNCTION: CategoryCompare
 *
 *  DESCRIPTION: Compare two categories
 *
 *  PARAMETERS:  attr1, attr2 - record attributes, which contain 
 *						  the category. 
 *					  appInfoH - MemHandle of the applications category info
 *
 *  RETURNS: 0 if they match, non-zero if not
 *				 + if s1 > s2
 *				 - if s1 < s2
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			art	8/5/96	Initial Revision
 *
 *************************************************************/
static Int16 CategoryCompare (UInt8 attr1, UInt8 attr2, MemHandle appInfoH)
{
	Int16 result;
	UInt8 category1;
	UInt8 category2;
	ToDoAppInfoPtr appInfoP;
	
	
	category1 = attr1 & dmRecAttrCategoryMask;
	category2 = attr2 & dmRecAttrCategoryMask;

	result = category1 - category2;
	if (result != 0)
		{
		if (category1 == dmUnfiledCategory)
			return (1);
		else if (category2 == dmUnfiledCategory)
			return (-1);
		
		appInfoP = (ToDoAppInfoPtr)MemHandleLock (appInfoH);
		result = StrCompare (appInfoP->categoryLabels[category1],
									appInfoP->categoryLabels[category2]);

		MemPtrUnlock (appInfoP);
		return result;
		}
	
	return result;
}


/************************************************************
 *
 *  FUNCTION: ToDoCompareRecords
 *
 *  DESCRIPTION: Compare two records.
 *
 *  PARAMETERS: database record 1
 *				database record 2
 *
 *  RETURNS: -n if record one is less (n != 0)
 *			  n if record two is less
 *
 *  CREATED: 1/23/95 
 *
 *  BY: Roger Flores
 *
 *	COMMENTS:	Compare the two records key by key until
 *	there is a difference.  Return -n if r1 is less or n if r2
 *	is less.  A zero is never returned because if two records
 *	seem identical then their unique IDs are compared!
 *
 * This function accepts record data chunk pointers to avoid
 * requiring that the record be within the database.  This is
 * important when adding records to a database.  This prevents
 * determining if a record is a deleted record (which are kept
 * at the end of the database and should be considered "greater").
 * The caller should test for deleted records before calling this
 * function!
 *
 *************************************************************/
static Int16 ToDoCompareRecords(ToDoDBRecordPtr r1, 
	ToDoDBRecordPtr r2, Int16 sortOrder, 
	SortRecordInfoPtr info1, SortRecordInfoPtr info2,
	MemHandle appInfoH)
{
	Int16 result;

	// Sort by priority, due date, and category.
	if (sortOrder == soPriorityDueDate)
		{
		result = (r1->priority & priorityOnly) - (r2->priority & priorityOnly);
		if (result == 0)
			{
			result = DateTypeCmp (r1->dueDate, r2->dueDate);
			if (result == 0)
				{
				result = CategoryCompare (info1->attributes, info2->attributes, appInfoH);
				}
			}
		}

	// Sort by due date, priority, and category.
	else if (sortOrder == soDueDatePriority)
		{
		result = DateTypeCmp(r1->dueDate, r2->dueDate);
		if (result == 0)
			{
			result = (r1->priority & priorityOnly) - (r2->priority & priorityOnly);
			if (result == 0)
				{
				result = CategoryCompare (info1->attributes, info2->attributes, appInfoH);
				}
			}
		}
	
	// Sort by category, priority, due date
	else if (sortOrder == soCategoryPriority)
		{
		result = CategoryCompare (info1->attributes, info2->attributes, appInfoH);
		if (result == 0)
			{
			result = (r1->priority & priorityOnly) - (r2->priority & priorityOnly);
			if (result == 0)
				{
				result = DateTypeCmp (r1->dueDate, r2->dueDate);
				}
			}
		}

	// Sort by category, due date, priority
	else if (sortOrder == soCategoryDueDate)
		{
		result = CategoryCompare (info1->attributes, info2->attributes, appInfoH);
		if (result == 0)
			{
			result = DateTypeCmp (r1->dueDate, r2->dueDate);
			if (result == 0)
				{
				result = (r1->priority & priorityOnly) - (r2->priority & priorityOnly);
				}
			}
		}
	
	return result;
}


/************************************************************
 *
 *  FUNCTION: ToDoGetAppInfo
 *
 *  DESCRIPTION: Get the app info chunk 
 *
 *  PARAMETERS: database pointer
 *
 *  RETURNS: MemHandle to the to do application info block (ToDoAppInfoType)
 *
 *  CREATED: 5/12/95 
 *
 *  BY: Art Lamb
 *
 *************************************************************/
static MemHandle ToDoGetAppInfo (DmOpenRef dbP)
{
	Err error;
	UInt16 cardNo;
	LocalID dbID;
	LocalID appInfoID;
	
	error = DmOpenDatabaseInfo (dbP, &dbID, NULL, NULL, &cardNo, NULL);
	ErrFatalDisplayIf (error,  "Get getting to do app info block");

	error = DmDatabaseInfo (cardNo, dbID, NULL, NULL, NULL, NULL, NULL, 
			NULL, NULL, &appInfoID, NULL, NULL, NULL);
	ErrFatalDisplayIf (error,  "Get getting to do app info block");

	return ((MemHandle) MemLocalIDToGlobal (appInfoID, cardNo));
}


/************************************************************
 *
 *  FUNCTION: ToDoGetSortOrder
 *
 *  DESCRIPTION: This routine gets the sort order value from the 
 *               'to do' application info block.
 *
 *  PARAMETERS: database pointer
 *
 *  RETURNS:    true - if the 'to do' record are sorted by priority, 
 *              false - if the records are sorted by due date.
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			art	5/12/95	Initial Revision
 *       art	3/22/96	Rename routine and added more sort orders
 *
 *************************************************************/
static UInt8 ToDoGetSortOrder (DmOpenRef dbP)
{
	UInt8 sortOrder;
	ToDoAppInfoPtr appInfoP;
			
	appInfoP = (ToDoAppInfoPtr)MemHandleLock (ToDoGetAppInfo (dbP));
	sortOrder = appInfoP->sortOrder;
	MemPtrUnlock (appInfoP);	

	return (sortOrder);
}


/************************************************************
 *
 *  FUNCTION: ToDoFindSortPosition
 *
 *  DESCRIPTION: Return where a record is or should be.
 *		Useful to find a record or find where to insert a record.
 *
 *  PARAMETERS: database record (not deleted!)
 *
 *  RETURNS: the size in bytes
 *
 *  CREATED: 1/11/95 
 *
 *  BY: Roger Flores
 *
 *************************************************************/
static UInt16 ToDoFindSortPosition(DmOpenRef dbP, ToDoDBRecord *newRecord, 
		SortRecordInfoPtr newRecordInfo)
{
	int sortOrder;
	
	sortOrder = ToDoGetSortOrder (dbP);
		
	return (DmFindSortPosition (dbP, newRecord, newRecordInfo, 
		(DmComparF *)ToDoCompareRecords, sortOrder));
}


/************************************************************
 *
 *  FUNCTION: ToDoNewRecord
 *
 *  DESCRIPTION: Create a new record in sorted position
 *
 *  PARAMETERS: database pointer
 *				database record
 *
 *  RETURNS: ##0 if successful, errorcode if not
 *
 *  CREATED: 1/10/95 
 *
 *  BY: Roger Flores
 *
 *************************************************************/
Err ToDoNewRecord(DmOpenRef dbP, ToDoItemPtr item, UInt16 category, UInt16 *index)
{
	Err 						err;
	UInt16					size;
	Char						zero=0;
	UInt16					attr;
	UInt16 					newIndex;
	UInt32					offset;
	ToDoDBRecordPtr		nilP=0;
	ToDoDBRecordPtr		recordP;
	MemHandle 				recordH;
	SortRecordInfoType	sortInfo;

	// Compute the size of the new to do record.
	size = sizeDBRecord;
	if (item->description)
		size += StrLen (item->description);
	if (item->note)
		size += StrLen (item->note);
		

	//  Allocate a chunk in the database for the new record.
	recordH = (MemHandle)DmNewHandle(dbP, size);
	if (recordH == NULL)
		return dmErrMemError;

	// Pack the the data into the new record.
	recordP = (ToDoDBRecordPtr)MemHandleLock (recordH);
	DmWrite(recordP, (UInt32)&nilP->dueDate, &item->dueDate, sizeof(nilP->dueDate));
	DmWrite(recordP, (UInt32)&nilP->priority, &item->priority, sizeof(nilP->priority));

	offset = (UInt32)&nilP->description;
	if (item->description)
		{
		DmStrCopy(recordP, offset, item->description);
		offset += StrLen (item->description) + 1;
		}
	else
		{
		DmWrite(recordP, offset, &zero, 1);
		offset++;
		}
	
	if (item->note) 
		DmStrCopy(recordP, offset, item->note);
	else
		DmWrite(recordP, offset, &zero, 1);
	
	
	sortInfo.attributes = category;
	sortInfo.uniqueID[0] = 0;
	sortInfo.uniqueID[1] = 0;
	sortInfo.uniqueID[2] = 0;
	
	// Determine the sort position of the new record.
	newIndex = ToDoFindSortPosition (dbP, recordP, &sortInfo);

	MemPtrUnlock (recordP);

	// Insert the record.
	err = DmAttachRecord(dbP, &newIndex, recordH, 0);
	if (err) 
		MemHandleFree(recordH);
	else
		{
		*index = newIndex;

		// Set the category.
		DmRecordInfo (dbP, newIndex, &attr, NULL, NULL);
		attr &= ~dmRecAttrCategoryMask;
		attr |= category;
		DmSetRecordInfo (dbP, newIndex, &attr, NULL);
		}
		
	return err;
}


#pragma mark ----
/***********************************************************************
 *
 * FUNCTION:    TasksDateTypeCmp
 *
 * DESCRIPTION: Compare two dates
 *
 * PARAMETERS:	database record 1
 *				database record 2
 *
 * RETURNS:		-n if record one is less (n != 0)
 *				n if record two is less
 *
 * HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		05/02/04	Initial Revision
 *
 ***********************************************************************/
static Int16 TasksDateTypeCmp (TasksDBRecordPtr r1, TasksDBRecordPtr r2)
{
	DateType	d1, d2;
	Int16		result = 0;


	if (r1->flag & 0xc0)
		d1 = *(DateType *)(&r1->description);
	else
		result = 1;
	if (r2->flag & 0xc0)
		d2 = *(DateType *)(&r2->description);
	else
		result += 2;
	switch (result)
	{
		case 1:
			return 1;
			break;
		case 2:
			return -1;
			break;
		case 3:
			return 0;
			break;
	}

	result = d1.year - d2.year;
	if (result != 0)
		return result;

	result = d1.month - d2.month;
	if (result != 0)
		return result;

	result = d1.day - d2.day;

	return result;
}


/***********************************************************************
 *
 * FUNCTION:    TasksCompareRecords
 *
 * DESCRIPTION: Compare two records
 *
 * PARAMETERS:	database record 1
 *				database record 2
 *
 * RETURNS:		-n if record one is less (n != 0)
 *				n if record two is less
 *
 * HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		05/02/04	Initial Revision
 *
 ***********************************************************************/
static Int16 TasksCompareRecords (TasksDBRecordPtr r1, TasksDBRecordPtr r2,
	 Int16 sortOrder, SortRecordInfoPtr info1, SortRecordInfoPtr info2,
	 MemHandle appInfoH)
{
	Int16			result;


	// Sort by priority, due date, and category.
	if (sortOrder == soPriorityDueDate)
	{
		result = r1->priority - r2->priority;
		if (result == 0)
		{
			result = TasksDateTypeCmp (r1, r2);
			if (result == 0)
			{
				result = CategoryCompare (info1->attributes, info2->attributes, appInfoH);
			}
		}
	}

	// Sort by due date, priority, and category.
	else if (sortOrder == soDueDatePriority)
	{
		result = TasksDateTypeCmp (r1, r2);
		if (result == 0)
		{
			result = r1->priority - r2->priority;
			if (result == 0)
			{
				result = CategoryCompare (info1->attributes, info2->attributes, appInfoH);
			}
		}
	}
	
	// Sort by category, priority, due date
	else if (sortOrder == soCategoryPriority)
	{
		result = CategoryCompare (info1->attributes, info2->attributes, appInfoH);
		if (result == 0)
		{
			result = r1->priority - r2->priority;
			if (result == 0)
			{
				result = TasksDateTypeCmp (r1, r2);
			}
		}
	}

	// Sort by category, due date, priority
	else if (sortOrder == soCategoryDueDate)
	{
		result = CategoryCompare (info1->attributes, info2->attributes, appInfoH);
		if (result == 0)
		{
			result = TasksDateTypeCmp (r1, r2);
			if (result == 0)
			{
				result = r1->priority - r2->priority;
			}
		}
	}
	
	return result;
}


/***********************************************************************
 *
 * FUNCTION:    TasksGetSortOrder
 *
 * DESCRIPTION:	This routine gets the sort order value from the 
 *				'to do' application info block.
 *
 * PARAMETERS:	database pointer
 *
 * RETURNS:		true - if the 'to do' record are sorted by priority, 
 *				false - if the records are sorted by due date.
 *
 * HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		05/02/04	Initial Revision
 *
 ***********************************************************************/
static UInt8 TasksGetSortOrder (DmOpenRef dbP)
{
	UInt8			sortOrder;
	TasksAppInfoPtr	appInfoP;

	appInfoP = (TasksAppInfoPtr)MemHandleLock (ToDoGetAppInfo (dbP));
	sortOrder = appInfoP->sortOrder;
	MemPtrUnlock (appInfoP);	

	return (sortOrder);
}


/***********************************************************************
 *
 * FUNCTION:    TasksFindSortPosition
 *
 * DESCRIPTION:	Return where a record is or should be. 
 *				'to do' application info block.
 *
 * PARAMETERS:	database record
 *
 * RETURNS:		the size in bytes
 *
 * HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		05/02/04	Initial Revision
 *
 ***********************************************************************/
static UInt16 TasksFindSortPosition (DmOpenRef dbP, TasksDBRecordPtr newRecord, 
		SortRecordInfoPtr newRecordInfo)
{
	int sortOrder;
	
	sortOrder = TasksGetSortOrder (dbP);

	return (DmFindSortPosition (dbP, newRecord, newRecordInfo, 
		(DmComparF *)TasksCompareRecords, sortOrder));
}


/************************************************************
 *
 * FUNCTION:	TasksNewRecord
 *
 * DESCRIPTION:	Create a new record in sorted position
 *
 * PARAMETERS:
 *
 * RETURNED:	errNone if successful, errorcode if not
 *
 * HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		05/02/04	Initial Revision
 *
 *************************************************************/
Err TasksNewRecord (DmOpenRef dbP, ToDoItemType *item, UInt16 category, UInt16 *index)
{
	Err 				err;
	UInt16				size;
	Char				zero=0;
	UInt16				attr;
	UInt16 				newIndex;
	UInt32				offset;
	TasksDBRecordType	*nilP=0;
	TasksDBRecordType	*recordP;
	MemHandle 			recordH;
	SortRecordInfoType	sortInfo;

	// Compute the size of the new to do record.
	size = sizeTasksDBRecord;
	if (item->description)
		size += StrLen (item->description);
	if (item->note)
		size += StrLen (item->note);


	//  Allocate a chunk in the database for the new record.
	recordH = (MemHandle)DmNewHandle(dbP, size);
	if (recordH == NULL)
		return dmErrMemError;

	// Pack the the data into the new record.
	recordP = (TasksDBRecordType *)MemHandleLock (recordH);
	DmSet (recordP, 0, size, 0);
	DmSet (recordP, (UInt32)&nilP->flag, sizeof(UInt8), 0x08);
	DmWrite (recordP, (UInt32)&nilP->priority, &item->priority, sizeof(nilP->priority));

	offset = (UInt32)&nilP->description;
	if (item->description)
	{
		DmStrCopy(recordP, offset, item->description);
		offset += StrLen (item->description) + 1;
	}
	else
	{
		DmWrite(recordP, offset, &zero, 1);
		offset++;
	}
	
	if (item->note) 
		DmStrCopy(recordP, offset, item->note);
	else
		DmWrite(recordP, offset, &zero, 1);


	sortInfo.attributes = category;
	sortInfo.uniqueID[0] = 0;
	sortInfo.uniqueID[1] = 0;
	sortInfo.uniqueID[2] = 0;

	// Determine the sort position of the new record.
	newIndex = TasksFindSortPosition (dbP, recordP, &sortInfo);

	MemPtrUnlock (recordP);

	// Insert the record.
	err = DmAttachRecord(dbP, &newIndex, recordH, 0);
	if (err) 
		MemHandleFree(recordH);
	else
	{
		*index = newIndex;

		// Set the category.
		DmRecordInfo (dbP, newIndex, &attr, NULL, NULL);
		attr &= ~dmRecAttrCategoryMask;
		attr |= category;
		DmSetRecordInfo (dbP, newIndex, &attr, NULL);
	}
		
	return err;
}


/***********************************************************************
 *
 * FUNCTION:    TasksGetRecord
 *
 * DESCRIPTION: This routine gets a record from the Tasks Database
 *
 * PARAMETERS:	recordH - MemHandle to be locked
 *				itemP - pointer to the ToDoItemType struct
 *
 * RETURNED:	errNone if successful, errorcode if not
 *
 * HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		05/02/04	Initial Revision
 *
 ***********************************************************************/
Err TasksGetRecord (MemHandle recordH, ToDoItemPtr itemP)
{
	TasksDBRecordPtr	recP;
	UInt16				offset = 0;


	if ((recP = (TasksDBRecordPtr)MemHandleLock (recordH)) == NULL)
		return dmErrIndexOutOfRange;

	itemP->priority = recP->priority | ((recP->flag & 0x40)? completeFlag : 0x00);
	if (recP->flag & 0xc0)
	{
		itemP->dueDate = *(DateType *)(&recP->description);
		offset += sizeof(DateType);
	}
	itemP->description = &recP->description + offset;
	offset += StrLen (itemP->description) + 1;
	itemP->note = &recP->description + offset;

	return errNone;
}


